/* ----------------------------------------------------------------------------
view_ = representing html grid
controller_ = for translating view to model
model_ = needed for 500811 command
---------------------------------------------------------------------------- */

const VIDEO_WIDTH = 1920;
const VIDEO_HEIGHT = 1080;

const CM_TO_INCH = 0.393701;

function alignBy8(x, a = 8)
{
    //var a = 8;
    var r = x % a;
    if (r >= (a / 2))
    {
        return x + (a - r);
    } else
    {
        return x - r;
    }
}

class MultiviewGrid
{
    constructor(rows, cols) 
    {
        this.viewRows = rows;
        this.viewCols = cols;
        this.viewDevices = [];
        this.viewTop;  // initial coords
        this.viewLeft;

        this.viewSizeX;
        this.viewSizeY;

        this.controllerGrid;

        this.modelTop = 0;  // initial coords for videowall/multiview in px,
        this.modelLeft = 0; // should always be (0,0)
        this.modelHeight;
        this.modelWidth;

        //virtual grid
        this.arrDisplayValues;
        this.arrSourceValues;
    }

    getDevice(mac)
    {
        var device = null;
        for (var key in this.viewDevices)
        {
            if (mac == this.viewDevices[key].mac) 
            {
                device = this.viewDevices[key];
                break;
            }
        }
        return device;
    }

    calculateWindows(doCompensations = true)
    {
        for (var key in this.viewDevices)
        {
            if (this.viewDevices[key] instanceof MultiviewDisplay)
            {
                var modelDisplay = this.viewDevices[key];

                if (!isNaN(modelDisplay.viewTop) && !isNaN(modelDisplay.viewLeft))
                {
                    this.calculateDisplayWindows(modelDisplay.mac, doCompensations);
                }
            }
        }
    }

    checkOverlap()
    {
        for (var key in this.viewDevices)
        {
            if (this.viewDevices[key] instanceof MultiviewDisplay)
            {
                var modelDisplay = this.viewDevices[key];

                if (!isNaN(modelDisplay.viewTop) && !isNaN(modelDisplay.viewLeft))
                {
                    this.calculateDisplayWindows(modelDisplay.mac);
                }
            }
        }
    }

    calculateDisplayWindows(macRX,doCompensations = true)
    {
        //var modelDisplay = this.viewDevices[macRX];
        var modelDisplay = this.getDevice(macRX);
        modelDisplay.removeAllWindows();

        for (var key in this.viewDevices)
        {
            if (this.viewDevices[key] instanceof MultiviewSource)
            {
                var modelSource = this.viewDevices[key];

                if (!isNaN(modelSource.viewTop) && !isNaN(modelSource.viewLeft))
                {
                    var window = this.generateWindow(modelDisplay, modelSource, doCompensations);
                    if (window instanceof MultiviewWindow) modelDisplay.addWindow(window);
                }
            }
        }
    }

    calculate_width_src(display, source)
    {
        return this.calculate_src('x', 'size', display, source);
    }

    calculate_height_src(display, source)
    {
        return this.calculate_src('y', 'size', display, source);
    }

    calculate_left_src(display, source)
    {
        return this.calculate_src('x', 'start', display, source);
    }

    calculate_top_src(display, source)
    {
        return this.calculate_src('y', 'start', display, source);
    }

    calculate_src(axis = 'x', variable = 'size', display, source)
    {
        if ((axis != 'x') && (axis != 'y') && (variable != 'start') && (variable != 'size'))
            return false;

        var newDisplay;

        if (axis == 'x')
        {
            var num_pre_displays = Math.abs(Math.floor((source.viewLeft - display.viewLeft) / display.viewWidth));
            var num_post_displays = Math.abs(Math.floor(((display.viewLeft + display.viewWidth) - (source.viewLeft + source.viewWidth)) / display.viewWidth));
        }
        else
        {
            var num_pre_displays = Math.abs(Math.floor((source.viewTop - display.viewTop) / display.viewHeight));
            var num_post_displays = Math.abs(Math.floor(((display.viewTop + display.viewHeight) - (source.viewTop + source.viewHeight)) / display.viewHeight));
        }

        var num_displays = num_pre_displays + 1 + num_post_displays;
        var current_key = num_pre_displays + 1;

        // initialize array
        if (typeof this.arrDisplayValues == 'undefined')
        {
            this.arrDisplayValues = [];
            this.arrSourceValues = [];

            for (var key = 1; key <= num_displays; key++) 
            {
                newDisplay = new MultiviewDisplay('virtual'); //mac = key                

                //copy values from current display
                newDisplay.bezelTop = display.bezelTop;
                newDisplay.bezelBottom = display.bezelBottom;
                newDisplay.bezelLeft = display.bezelLeft;
                newDisplay.bezelRight = display.bezelRight;

                newDisplay.screenSize = display.screenSize;
                newDisplay.screenWidth = display.screenWidth;
                newDisplay.screenHeight = display.screenHeight;
                newDisplay.measureUnit = display.measureUnit;

                newDisplay.viewLeft = display.viewLeft;
                newDisplay.viewTop = display.viewTop;
                newDisplay.viewWidth = display.viewWidth;
                newDisplay.viewHeight = display.viewHeight;

                //newDisplay.viewLeft = display.viewLeft;
                //newDisplay.viewTop = display.viewTop;

                if (key <= num_pre_displays)
                {
                    // add virtual pre displays
                    if (axis == 'x') newDisplay.viewLeft = parseInt(display.viewLeft) - (key * parseInt(display.viewWidth));
                    if (axis == 'y') newDisplay.viewTop = parseInt(display.viewTop) - (key * parseInt(display.viewHeight));

                    this.arrDisplayValues[key] = newDisplay;
                    this.arrSourceValues[key] = new MultiviewSourceValues();
                }
                else if (key == current_key)
                {
                    // add current display
                    this.arrDisplayValues[key] = display;
                    this.arrSourceValues[key] = new MultiviewSourceValues();
                }
                else if ((key > current_key) /*&& (key < current_key + num_post_displays))*/)
                {
                    // add virtual post displays                    
                    if (axis == 'x') newDisplay.viewLeft = parseInt(display.viewLeft) + ((key - current_key) * parseInt(display.viewWidth));
                    if (axis == 'y') newDisplay.viewTop = parseInt(display.viewTop) + ((key - current_key) * parseInt(display.viewHeight));

                    this.arrDisplayValues[key] = newDisplay;
                    this.arrSourceValues[key] = new MultiviewSourceValues();
                }
            }
        }
        else
        {
            this.arrDisplayValues[current_key] = display;
            this.arrSourceValues[current_key] = new MultiviewSourceValues();
        }

        var total_size_inches = 0;
        var total_bezel_inches = 0;

        var newCoords;
        var displayCoords;

        //get totals
        for (var key = 1; key <= num_displays; key++) 
        {
            newCoords = new MultiviewCoordinates(this.arrDisplayValues[key], source);

            if (axis == 'x') 
            {
                total_size_inches += parseFloat(this.arrDisplayValues[key].screenWidth * (newCoords.windowWidth / display.viewWidth)); // screen width * ratio                

                if (key > 1) // ignore first display
                    total_bezel_inches += parseFloat(this.arrDisplayValues[key].bezelLeft) + parseFloat(this.arrDisplayValues[key].bezelRight);
            }
            else
            {
                total_size_inches += parseFloat(this.arrDisplayValues[key].screenHeight * (newCoords.windowHeight / display.viewHeight)); // screen width * ratio                

                if (key > 1) // ignore first display
                    total_bezel_inches += parseFloat(this.arrDisplayValues[key].bezelTop) + parseFloat(this.arrDisplayValues[key].bezelBottom);
            }
        }

        var display_size_src;

        var sum_inches = 0;
        var partial_sum_px = 0;

        // size coordinates *************************************************

        for (var key = 1; key <= current_key; key++)
        {
            displayCoords = new MultiviewCoordinates(this.arrDisplayValues[key], source);

            if (axis == 'x')
                display_size_src = ((this.arrDisplayValues[key].screenWidth * (displayCoords.windowWidth / this.arrDisplayValues[key].viewWidth)) / (total_size_inches + total_bezel_inches)) * source.res_width;
            else
                display_size_src = ((this.arrDisplayValues[key].screenHeight * (displayCoords.windowHeight / this.arrDisplayValues[key].viewHeight)) / (total_size_inches + total_bezel_inches)) * source.res_height;

            if (axis == 'x') //adjustments, only on x
            {
                newCoords = new MultiviewCoordinates(this.arrDisplayValues[key], source);

                sum_inches += parseFloat(this.arrDisplayValues[key].screenWidth * (newCoords.windowWidth / display.viewWidth)); // screen width * ratio

                if (key > 1)
                {
                    var compensated_bezel_px = alignBy8(((parseFloat(this.arrDisplayValues[key].bezelLeft) + parseFloat(this.arrDisplayValues[key].bezelRight)) / (total_size_inches + total_bezel_inches)) * source.res_width);

                    num_pre_displays = key - 1;

                    var total_width_inches = sum_inches + (((parseFloat(display.bezelLeft) + parseFloat(display.bezelRight))) * num_pre_displays);
                    var total_width_px = (total_width_inches / (total_size_inches + total_bezel_inches)) * source.res_width;
                    var new_width_px = total_width_px - (compensated_bezel_px * num_pre_displays) - partial_sum_px;

                    display_size_src = alignBy8(new_width_px);
                }
            }

            //save the value to the array
            if (axis == 'x') this.arrSourceValues[key].window_width_src = display_size_src;
            if (axis == 'y') this.arrSourceValues[key].window_height_src = display_size_src;

            if (axis == 'x') // part of adjustments
            {
                if (key >= 1)
                {
                    partial_sum_px += alignBy8(this.arrSourceValues[key].window_width_src);
                }
            }
        }

        // start coordinates *************************************************
        var start_sum = 0;
        var start_bezel = 0;

        var start_bezel_inches = 0;

        for (var key = 1; key <= current_key; key++)
        {
            if (key > 1)
            {
                if (axis == 'x')
                {
                    start_sum += this.arrSourceValues[key - 1].window_width_src;
                    start_bezel_inches += parseFloat(this.arrDisplayValues[key - 1].bezelLeft) + parseFloat(this.arrDisplayValues[key - 1].bezelRight);
                }
                else
                {
                    start_sum += this.arrSourceValues[key - 1].window_height_src;
                    start_bezel_inches += parseFloat(this.arrDisplayValues[key - 1].bezelTop) + parseFloat(this.arrDisplayValues[key - 1].bezelBottom);
                }
            }

            //convert bezel inches to pixels
            if (axis == 'x')
                start_bezel = ((start_bezel_inches) / (total_size_inches + total_bezel_inches)) * source.res_width;
            else
                start_bezel = ((start_bezel_inches) / (total_size_inches + total_bezel_inches)) * source.res_height;

            if (axis == 'x')  // adjustments, only on x
            {
                var compensated_bezel_px = alignBy8(((parseFloat(this.arrDisplayValues[key].bezelLeft) + parseFloat(this.arrDisplayValues[key].bezelRight)) / (total_size_inches + total_bezel_inches)) * source.res_width);

                //from size compensations
                if (key > 1) start_bezel = (compensated_bezel_px * (key - 1));
            }

            if (axis == 'x') this.arrSourceValues[key].window_xcoord_src = start_sum + start_bezel;
            if (axis == 'y') this.arrSourceValues[key].window_ycoord_src = start_sum + start_bezel;
        }

        if (axis == 'x') 
        {
            if (variable == 'size') return this.arrSourceValues[current_key].window_width_src;
            else return this.arrSourceValues[current_key].window_xcoord_src;
        }
        if (axis == 'y') 
        {
            if (variable == 'size') return this.arrSourceValues[current_key].window_height_src;
            else return this.arrSourceValues[current_key].window_ycoord_src;
        }
    }

    generateWindow(display, source, isCompensating = true)
    {
        var VIDEO_WIDTH_DST = VIDEO_WIDTH;
        var VIDEO_HEIGHT_DST = VIDEO_HEIGHT;

        var VIDEO_WIDTH_SRC = source.res_width;
        var VIDEO_HEIGHT_SRC = source.res_height;

        var generateCoords = new MultiviewCoordinates(display, source);

        if (generateCoords.isGood() === false) return false;

        var windowX = (generateCoords.windowX / display.viewWidth) * VIDEO_WIDTH_DST;
        var windowY = (generateCoords.windowY / display.viewHeight) * VIDEO_HEIGHT_DST;
        var windowWidth = (generateCoords.windowWidth / display.viewWidth) * VIDEO_WIDTH_DST;
        var windowHeight = (generateCoords.windowHeight / display.viewHeight) * VIDEO_HEIGHT_DST;

        var windowXRatio = generateCoords.windowXRatio;
        var windowYRatio = generateCoords.windowYRatio;
        var windowWidthRatio = generateCoords.windowWidthRatio;
        var windowHeightRatio = generateCoords.windowHeightRatio;

        var window_xcoord_src = windowXRatio * VIDEO_WIDTH_SRC;
        var window_ycoord_src = windowYRatio * VIDEO_HEIGHT_SRC;
        var window_width_src = windowWidthRatio * VIDEO_WIDTH_SRC;
        var window_height_src = windowHeightRatio * VIDEO_HEIGHT_SRC;

        /*
        console.log("*** Pre-Bezel:  **************************");
        console.log("windowX: " + windowX);
        console.log("windowY: " + windowY);
        console.log("windowWidth: " + windowWidth);
        console.log("windowHeight: " + windowHeight);
        console.log("***************************");
        console.log("windowXRatio: " + windowXRatio);
        console.log("windowYRatio: " + windowYRatio);
        console.log("windowWidthRatio: " + windowWidthRatio);
        console.log("windowHeightRatio: " + windowHeightRatio);
        console.log("***************************");
        console.log("window_xcoord_src: " + window_xcoord_src);
        console.log("window_ycoord_src: " + window_ycoord_src);
        console.log("window_width_src: " + window_width_src);
        console.log("window_height_src: " + window_height_src);
        console.log("******************************************");
        */

        // Shift and scale ****************************************************        

        var pixelShift = 8;
        var pixelScale = 8;

        //Bezel calculations ***************************************************

        var isBezeling = true;
        if (isBezeling === true)
        {
            window_width_src = this.calculate_width_src(display, source);
            window_xcoord_src = this.calculate_left_src(display, source);

            //clear virtual grid
            this.arrDisplayValues = undefined;
            this.arrSourceValues = undefined;

            window_height_src = this.calculate_height_src(display, source);
            window_ycoord_src = this.calculate_top_src(display, source);

            //clear virtual grid
            this.arrDisplayValues = undefined;
            this.arrSourceValues = undefined;

            /*
            console.log("*** Post-Bezel:  **************************");
            console.log("window_xcoord_src: " + window_xcoord_src);
            console.log("window_ycoord_src: " + window_ycoord_src);
            console.log("window_width_src: " + window_width_src);
            console.log("window_height_src: " + window_height_src);
            console.log("******************************************");
            */
        }

        // Source Resolution Conversion ***************************************

        /*
        window_width_src = window_width_src * (VIDEO_WIDTH_SRC / 1920);
        window_xcoord_src = window_xcoord_src * (VIDEO_WIDTH_SRC / 1920);
        window_height_src = window_height_src * (VIDEO_HEIGHT_SRC / 1080);
        window_ycoord_src = window_ycoord_src * (VIDEO_HEIGHT_SRC / 1080);
        */

        // Compensations ******************************************************                

        var pixelModding = 8;

        //var isCompensating = true;
        if (isCompensating === true)
        {
            //display
            var windowX2 = windowX + windowWidth;
            var windowY2 = windowY + windowHeight;

            windowX = alignBy8(windowX);
            windowWidth = alignBy8(windowX2) - windowX;

            windowY = alignBy8(windowY);
            windowHeight = alignBy8(windowY2) - windowY;

            //source            
            var window_xcoord_src2 = window_xcoord_src + window_width_src;
            var window_ycoord_src2 = window_ycoord_src + window_height_src;

            window_xcoord_src = alignBy8(window_xcoord_src);
            window_width_src = alignBy8(window_xcoord_src2) - window_xcoord_src;

            window_ycoord_src = alignBy8(window_ycoord_src);
            window_height_src = alignBy8(window_ycoord_src2) - window_ycoord_src;
        }

        // ********************************************************************

        /*
        console.log("*******" + "window_xcoord_src: " + window_xcoord_src + "*******");
        console.log("*******" + "window_ycoord_src: " + window_ycoord_src + "*******");
        console.log("*******" + "window_width_src: " + window_width_src + "*******");
        console.log("*******" + "window_height_src: " + window_height_src + "*******");
        */

        var window = new MultiviewWindow(windowX, windowY, windowHeight, windowWidth,
            windowXRatio, windowYRatio, windowHeightRatio, windowWidthRatio,
            window_xcoord_src, window_ycoord_src, window_height_src, window_width_src);

        window.num_aud_on = source.num_aud_on;
        window.zorder = source.zorder;
        window.aspectCvrs = source.aspectCvrs;
        window.repeatable = source.repeatable;
        window.low_delay = source.low_delay;
        window.src_onplaying = source.modelSourceURL;

        return window;
    }

    isSingleDisplay()
    {
        var displayCount=0;
        var macAddress;

        var device;
        for (var key in this.viewDevices)
        {
            device = this.viewDevices[key];

            if (!isNaN(device.viewTop) && !isNaN(device.viewLeft))
            {
                if (device.type == 'display')
                {
                    displayCount++;
                    macAddress = device.mac;
                }
            }
            if (displayCount > 1) return false;
        }
        
        if (displayCount == 0) return false;
        else return macAddress;
    }

    verifyControllerGrid()
    {
        var tempGrid = new Array(this.viewRows)
        for (var i = 0; i < this.viewRows; i++) 
        {
            tempGrid[i] = new Array(this.viewCols);
        }

        var device;
        for (var key in this.viewDevices)
        {
            device = this.viewDevices[key];

            if (!isNaN(device.viewTop) && !isNaN(device.viewLeft))
            {
                if (device.type == 'display' || device.type == 'source')
                {
                    for (var i = device.viewTop; i < (device.viewTop + device.viewHeight); i++)
                    {
                        for (var j = device.viewLeft; j < (device.viewLeft + device.viewWidth); j++)
                        {
                            if ((i < this.viewRows) && (j < this.viewCols))
                            {
                                if (typeof tempGrid[i][j] == 'undefined')
                                {
                                    tempGrid[i][j] = new Array();
                                    tempGrid[i][j]['type'] = device.type;
                                }
                                else if (device.type == 'source' && (tempGrid[i][j]['type'] == 'source+source' || tempGrid[i][j]['type'] == 'both+source'))
                                {
                                    return "OVERLAPPING SOURCES";// source+source+source overlap
                                }
                                else if (tempGrid[i][j]['type'] != device.type && tempGrid[i][j]['type'] != 'both' && tempGrid[i][j]['type'] != 'both+source') // you can overlap display/source, but not display/display or source/source
                                {
                                    tempGrid[i][j]['type'] = 'both';
                                }
                                else if (device.type == 'source' && (tempGrid[i][j]['type'] == 'source' || tempGrid[i][j]['type'] == 'both'))  //ignore source ovelap
                                {
                                    tempGrid[i][j]['type'] = tempGrid[i][j]['type'].toString() + '+source';
                                }
                                else
                                {
                                    return "INCORRECT PLACEMENT"; //overlap
                                }
                            }
                            else
                            {
                                return "OUT OF BOUNDS"; //out of bounds
                            }
                        }
                    }
                }
            }
        }

        //check size
        this.findInitialCoords();
        for (var keyRX in this.viewDevices)      
        {
            if (this.viewDevices[keyRX] instanceof MultiviewDisplay)
            {
                var modelDisplay = this.viewDevices[keyRX];
                if (!isNaN(modelDisplay.viewTop) && !isNaN(modelDisplay.viewLeft))
                {
                    for (var keyTX in this.viewDevices)
                    {
                        if (this.viewDevices[keyTX] instanceof MultiviewSource)
                        {
                            var modelSource = this.viewDevices[keyTX];
                            if (!isNaN(modelSource.viewTop) && !isNaN(modelSource.viewLeft))
                            {
                                var window = this.generateWindow(modelDisplay, modelSource);
                                if (window instanceof MultiviewWindow) 
                                {                                    
                                    //if ((window.width < 144) || (window.height < 144) || (window.window_width_src < 144) || (window.window_height_src < 144))                                    
                                    if ((window.width < 144) || (window.height < 144))
                                        return "SOURCE TOO SMALL"; //too small                                        
                                }
                            }
                        }
                    }
                }
            }
        }

        //console.log(tempGrid);
        return true;

    }

    addDevice(device)
    {
        if (device instanceof MultiviewDevice)
        {
            //this.viewDevices[device.mac] = device;
            return this.viewDevices.push(device);
        }
        else return false;
    }

    removeDevice(index)
    {
        this.viewDevices.splice(index, 1);
    }

    removeAllDevices()
    {
        this.viewDevices.splice(0, this.viewDevices.length); // or this.viewDevices = [];
        this.viewTop;
        this.viewLeft;
    }

    findInitialCoords()
    {
        var data = this.sortDevices('viewTop');
        this.viewTop = parseInt(data[0].viewTop);
        var data = this.sortDevices('viewLeft');
        this.viewLeft = parseInt(data[0].viewLeft);
    }

    findSize()
    {
        this.viewSizeX = 0;
        this.viewSizeY = 0;

        this.findInitialCoords();
        for (var key in this.viewDevices)
        {
            if (!isNaN(this.viewDevices[key].viewTop) && !isNaN(this.viewDevices[key].viewLeft))
            {
                this.viewSizeX = Math.max(this.viewSizeX, ((this.viewDevices[key].viewLeft - this.viewLeft) + this.viewDevices[key].viewWidth));
                this.viewSizeY = Math.max(this.viewSizeY, ((this.viewDevices[key].viewTop - this.viewTop) + this.viewDevices[key].viewHeight));
            }
        }
    }

    sortDevices(parameter)
    {
        var data = [];
        var i = 0;
        for (var key in this.viewDevices)
        {
            data[i] = {};
            $.extend(true, data[i], this.viewDevices[key]);
            i++;
        }

        data.sort((function (index)
        {
            return function (a, b)
            {
                if (isNaN(a[index])) return 1;
                if (isNaN(b[index])) return -1;
                if (a[index] < b[index]) return -1;
                if (a[index] > b[index]) return 1;
                return 0;
            };
        })(parameter));

        return data;
    }

    printSize()
    {
        this.findSize();
        console.log('Size: (' + ((isNaN(this.viewSizeX)) ? "none" : this.viewSizeX.toString()) + ',' + ((isNaN(this.viewSizeY)) ? "none" : this.viewSizeY.toString()) + ')');
    }

    printInitialCoords()
    {
        this.printInitialCoords();
        console.log('Initial Point: (' + ((isNaN(this.viewLeft)) ? "none" : this.viewLeft.toString()) + ',' + ((isNaN(this.viewTop)) ? "none" : this.viewTop.toString()) + ')');
    }
}

class MultiviewWindow
{
    constructor(x = null, y = null,
        height = null, width = null,
        x_ratio = "0.000000", y_ratio = "0.000000",
        height_ratio = "0.000000", width_ratio = "0.000000",
        window_xcoord_src = "0", window_ycoord_src = "0",
        window_height_src = "0", window_width_src = "0")
    {
        this.x; this.y;
        this.height; this.width;

        this.x_ratio = "0.000000"; this.y_ratio = "0.000000";
        this.height_ratio = "0.000000"; this.width_ratio = "0.000000";

        this.window_xcoord_src = "0";
        this.window_ycoord_src = "0";
        this.window_width_src = "0";
        this.window_height_src = "0";

        this.src_onplaying = null;

        this.fill_attributes(x, y, height, width, x_ratio, y_ratio, height_ratio, width_ratio, window_xcoord_src, window_ycoord_src, window_height_src, window_width_src);

        this.num_aud_on = 0;
        this.zorder = 2;
        this.aspectCvrs = 0;
        this.low_delay = 0;
        this.rotate = 0;
        this.repeatable = 1;
    }

    getArray(old = false)
    {
        var data = [];

        //"window_zorder":"0","window_aspectCvrs":"1","window_rotate":"0"

        data['num_aud_on'] = parseInt(this.num_aud_on).toString();
        data['window_zorder'] = parseInt(this.zorder).toString();
        data['window_aspectCvrs'] = parseInt(this.aspectCvrs).toString();
        data['window_rotate'] = parseInt(this.rotate).toString();
        data['low_delay'] = parseInt(this.low_delay).toString();
        data['repeatable'] = parseInt(this.repeatable).toString();            

        data['window_xcoord_dst'] = parseInt(this.x).toString();
        data['window_ycoord_dst'] = parseInt(this.y).toString();
        data['window_width_dst'] = parseInt(Math.ceil(this.width)).toString();
        data['window_height_dst'] = parseInt(Math.ceil(this.height)).toString();

        if (old == true)
        {
            data['window_xcoord_ratio'] = parseFloat(this.x_ratio).toPrecision(6).toString();
            data['window_ycoord_ratio'] = parseFloat(this.y_ratio).toPrecision(6).toString();
            data['window_width_ratio'] = parseFloat(this.width_ratio).toPrecision(6).toString();
            data['window_height_ratio'] = parseFloat(this.height_ratio).toPrecision(6).toString();
        }
        else
        {
            data['window_xcoord_src'] = parseInt(this.window_xcoord_src).toString();
            data['window_ycoord_src'] = parseInt(this.window_ycoord_src).toString();
            data['window_width_src'] = parseInt(this.window_width_src).toString();
            data['window_height_src'] = parseInt(this.window_height_src).toString();
        }

        data['src_onplaying'] = this.src_onplaying.toString();
        //data['src_onplaying'] = '';

        return data;
    }

    fill_attributes(x = this.x, y = this.y, height = this.height, width = this.width,
        x_ratio = this.x_ratio, y_ratio = this.y_ratio, height_ratio = this.height_ratio, width_ratio = this.width_ratio,
        window_xcoord_src = this.window_xcoord_src, window_ycoord_src = this.window_ycoord_src, window_height_src = this.window_height_src, window_width_src = this.window_width_src)
    {
        this.x = Math.ceil(x).toString();
        this.y = Math.ceil(y).toString();
        this.height = Math.ceil(height).toString();
        this.width = Math.ceil(width).toString();
        this.x_ratio = parseFloat(x_ratio).toPrecision(6).toString();
        this.y_ratio = parseFloat(y_ratio).toPrecision(6).toString();
        this.height_ratio = parseFloat(height_ratio).toPrecision(6).toString();
        this.width_ratio = parseFloat(width_ratio).toPrecision(6).toString();
        this.window_xcoord_src = parseInt(window_xcoord_src).toString();
        this.window_ycoord_src = parseInt(window_ycoord_src).toString();
        this.window_width_src = parseInt(window_width_src).toString();
        this.window_height_src = parseInt(window_height_src).toString();
    }

    fill_source(url = null)
    {
        this.src_onplaying = url;
    }
}

class MultiviewDevice
{
    constructor(mac, type)
    {
        this.mac = mac;
        this.type = type;

        this.viewTop;
        this.viewLeft;
        this.viewHeight = 1;
        this.viewWidth = 1;

        this.modelTop;
        this.modelLeft;
        this.modelHeight;
        this.modelWidth;
    }

    isInGrid()
    {
        if (!isNaN(this.viewTop) && !isNaN(this.viewLeft)) return true; else return false;
    }

    modifyCoord(x = null, y = null)
    {
        this.viewLeft = (x === null) ? null : parseInt(x);
        this.viewTop = (y === null) ? null : parseInt(y);
    }

    printCoord()
    {
        console.log('Device Coords: (' + ((isNaN(this.viewLeft)) ? "false" : this.viewLeft.toString()) + ',' + ((isNaN(this.viewTop)) ? "false" : this.viewTop.toString()) + ')');
    }

    modifySize(width, height)
    {
        this.viewHeight = parseInt(height);
        this.viewWidth = parseInt(width);
        //this.printSize();
    }

    printSize()
    {
        console.log('Device Size: (' + ((isNaN(this.viewHeight)) ? "false" : this.viewHeight.toString()) + 'x' + ((isNaN(this.viewWidth)) ? "false" : this.viewWidth.toString()) + ')');
    }

}

class MultiviewDisplay extends MultiviewDevice
{
    constructor(mac = null)
    {
        super(mac, 'display');

        this.modelWidth = VIDEO_WIDTH; //assume all displays as VIDEO_HEIGHT
        this.modelHeight = VIDEO_HEIGHT;

        this.modelWindows = [];

        this.display_mode = 1;
        this.display_res = 100;
        this.deep_color = 0;
        this.num_aud_on = 0;

        this.bezelTop = 0.0;
        this.bezelBottom = 0.0;
        this.bezelLeft = 0.0;
        this.bezelRight = 0.0;

        this.screenSize = 21; // in inches
        this.screenWidth = 18.303;
        this.screenHeight = 10.2955;
        this.measureUnit = "inch";
    }

    getScreenSize()
    {
        //this.screenSize = Math.round(Math.sqrt((this.screenWidth * this.screenWidth) + (this.screenHeight * this.screenHeight)));
        this.screenSize = Math.sqrt((this.screenWidth * this.screenWidth) + (this.screenHeight * this.screenHeight)).toFixed(1);
        return (this.screenSize)
    }


    setScreenSize(screenSize)
    {
        // 16x9 angle = 29.37
        // Cos(29.37) = 0.87    and   Sin(29.37) = 0.47 

        this.screenSize = parseFloat(screenSize);
        this.screenWidth = 0.87157 * parseFloat(this.screenSize);
        this.screenHeight = 0.49026 * parseFloat(this.screenSize);
    }

    getArray()
    {
        var data = [];

        data['macRx'] = this.mac.toString();

        data['display_mode'] = parseInt(this.display_mode).toString();
        data['display_res'] = parseInt(this.display_res).toString();
        data['deep_color'] = parseInt(this.deep_color).toString();
        data['num_aud_on'] = parseInt(this.num_aud_on).toString();

        data['mvCoordX'] = parseInt(this.viewLeft);
        data['mvCoordY'] = parseInt(this.viewTop);
        data['mvSizeX'] = parseInt(this.viewWidth);
        data['mvSizeY'] = parseInt(this.viewHeight);

        data['bezelTop'] = this.bezelTop;
        data['bezelBottom'] = this.bezelBottom;
        data['bezelLeft'] = this.bezelLeft;
        data['bezelRight'] = this.bezelRight;

        data['scaleUpH'] = 0;
        data['scaleDownH'] = 0;
        data['scaleUpV'] = 0;
        data['scaleDownV'] = 0;

        data['shiftUp'] = 0;
        data['shiftDown'] = 0;
        data['shiftRight'] = 0;
        data['shiftLeft'] = 0;

        data['screenWidth'] = this.screenWidth;
        data['screenHeight'] = this.screenHeight;
        data['measureUnit'] = this.measureUnit;

        return data;
    }

    hasWindows()
    {
        return !(this.modelWindows.length === 0);
    }

    hasSource(url)
    {
        for (var key in this.modelWindows)
        {
            if (this.modelWindows[key].src_onplaying == url) return true;
        }
        return false;
    }

    removeAllWindows()
    {
        this.modelWindows.splice(0, this.modelWindows.length); // or this.modelWindows = [];
    }

    addWindow(x, y, height, width, x_ratio, y_ratio, height_ratio, width_ratio, x_src, y_src, height_src, width_src)
    {
        this.modelWindows.push(new MultiviewWindow(x, y, height, width, x_ratio, y_ratio, height_ratio, width_ratio, x_src, y_src, height_src, width_src));
        return true;
    }

    addWindow(window)
    {
        if (window instanceof MultiviewWindow)
        {
            this.modelWindows.push(window);
            return true;
        }
        else return false;
    }

    getWindow(url)
    {
        for (var key in this.modelWindows)
        {
            if (this.modelWindows[key].src_onplaying == url) return this.modelWindows[key];
        }
        return false;
    }

}

class MultiviewSource extends MultiviewDevice
{
    constructor(mac = null, url = null, resolution = "1920x1080")
    {
        super(mac, 'source');

        this.zorder = 2;
        this.aspectCvrs = 0;
        this.rotate = 0;
        this.repeatable = 1;
        this.low_delay = 1;
        this.num_aud_on = 0;

        this.modelSourceURL = url;

        this.shiftUp = 0;
        this.shiftDown = 0;
        this.shiftRight = 0;
        this.shiftLeft = 0;

        this.scaleUpH = 0;
        this.scaleDownH = 0;
        this.scaleUpV = 0;
        this.scaleDownV = 0;

        if (resolution == "") resolution = "1920x1080";

        this.res_width = parseInt(resolution.split('x')[0]);
        this.res_height = parseInt(resolution.split('x')[1]);
    }

    getArray()
    {
        var data = [];

        data['macTx'] = this.mac.toString();
        data['src_onplaying'] = this.modelSourceURL.toString();
        data['num_aud_on'] = this.num_aud_on;
        data['zorder'] = this.zorder;
        data['aspectCvrs'] = this.aspectCvrs;
        data['rotate'] = this.rotate;
        data['low_delay'] = this.low_delay;
        data['repeatable'] = this.repeatable;

        data['mvCoordX'] = parseInt(this.viewLeft);
        data['mvCoordY'] = parseInt(this.viewTop);
        data['mvSizeX'] = parseInt(this.viewWidth);
        data['mvSizeY'] = parseInt(this.viewHeight);

        data['scaleUpH'] = parseInt(this.scaleUpH);
        data['scaleDownH'] = parseInt(this.scaleDownH);
        data['scaleUpV'] = parseInt(this.scaleUpV);
        data['scaleDownV'] = parseInt(this.scaleDownV);

        data['shiftUp'] = parseInt(this.shiftUp);
        data['shiftDown'] = parseInt(this.shiftDown);
        data['shiftRight'] = parseInt(this.shiftRight);
        data['shiftLeft'] = parseInt(this.shiftLeft);

        data['screenWidth'] = 0;
        data['screenHeight'] = 0;
        data['measureUnit'] = "inch";

        return data;
    }
}

class MultiviewSourceValues
{
    constructor()
    {
        this.window_xcoord_src;
        this.window_ycoord_src;
        this.window_width_src;
        this.window_height_src;
    }
}


class MultiviewCoordinates
{
    constructor(display = false, source = false) 
    {
        this.windowViewWidth;
        this.windowViewHeight;
        this.windowViewX;
        this.windowViewY;

        this.windowX;
        this.windowY;
        this.windowWidth;
        this.windowHeight;

        this.bezelRight;
        this.bezelBottom;
        this.bezelLeft;
        this.bezelTop;

        // for old RXs
        this.windowXRatio;
        this.windowYRatio;
        this.windowWidthRatio;
        this.windowHeightRatio;

        this.hasWindow = false;

        if ((display !== false) && (source !== false))
        {
            this.calculate(display, source);
        }
    }

    isGood()
    {
        return (this.hasWindow);
    }

    print()
    {
        console.log("View: " + this.windowViewX + " " + this.windowViewY + " " + this.windowViewWidth + " " + this.windowViewHeight);
        console.log("Window: " + this.windowX + " " + this.windowY + " " + this.windowWidth + " " + this.windowHeight);
    }

    calculate(display, source)
    {
        var l1x = display.viewLeft;
        var l1y = display.viewTop;
        var r1x = display.viewLeft + display.viewWidth;
        var r1y = display.viewTop + display.viewHeight;

        var l2x = source.viewLeft;
        var l2y = source.viewTop;
        var r2x = source.viewLeft + source.viewWidth;
        var r2y = source.viewTop + source.viewHeight;

        this.windowViewWidth = Math.max(0, Math.min(r1x, r2x) - Math.max(l1x, l2x));
        this.windowViewHeight = Math.max(0, Math.min(r1y, r2y) - Math.max(l1y, l2y));

        // if source is not on display check
        if ((Math.max(0, Math.min(r1x, r2x) - Math.max(l1x, l2x)) * Math.max(0, Math.min(r1y, r2y) - Math.max(l1y, l2y)) === 0)) return false;


        // if source is not on display check
        if ((this.windowViewWidth * this.windowViewHeight) === 0) return false;

        this.windowViewX = Math.max(l1x, l2x);
        this.windowViewY = Math.max(l1y, l2y);

        this.windowX = parseInt(this.windowViewX - l1x);
        this.windowY = parseInt(this.windowViewY - l1y);
        this.windowWidth = parseInt(this.windowViewWidth);
        this.windowHeight = parseInt(this.windowViewHeight);

        // for old RXs
        this.windowXRatio = parseFloat(((this.windowViewX - l2x) / (r2x - l2x)));
        this.windowYRatio = parseFloat(((this.windowViewY - l2y) / (r2y - l2y)));
        this.windowWidthRatio = ((this.windowViewWidth / (r2x - l2x)));
        this.windowHeightRatio = ((this.windowViewHeight / (r2y - l2y)));

        this.hasWindow = true;

        /*
        //bezel calculations
        var TempSourceWidth = (r2x - l2x);
        var TempSourceHeight = (r2y - l2y);

        var windowXRatio = parseFloat(((this.windowViewX - l2x) / TempSourceWidth));
        var windowYRatio = parseFloat(((this.windowViewY - l2y) / TempSourceHeight));;

        var windowWidthRatio = ((this.windowViewWidth / TempSourceWidth));
        var windowHeightRatio = ((this.windowViewHeight / TempSourceHeight));

        this.bezelRight = ((windowXRatio + windowWidthRatio) < 0.99) ? ((display.bezelRight)) : 0;
        this.bezelBottom = ((windowYRatio + windowHeightRatio) < 0.99) ? ((display.bezelBottom)) : 0;
        this.bezelLeft = ((windowXRatio - windowWidthRatio) >= 0) ? ((display.bezelLeft)) : 0;
        this.bezelTop = ((windowYRatio - windowHeightRatio) >= 0) ? ((display.bezelTop)) : 0;

        if (display.measureUnit == "cm")
        {
            this.bezelRight = this.bezelRight * CM_TO_INCH;
            this.bezelBottom = this.bezelBottom * CM_TO_INCH;
            this.bezelLeft = this.bezelLeft * CM_TO_INCH;
            this.bezelTop = this.bezelTop * CM_TO_INCH;
        }

        //console.log("Bezel: " + this.bezelRight + " " + this.bezelBottom + " " + this.bezelLeft + " " + this.bezelTop)
        */
    }
}
